package dag;

#### We need 5.6.0 for 'our' variable definitions
require v5.6.0;
use strict;
use warnings;

use Data::Dumper;
use File::Find;
# use Text::Balanced qw/extract_codeblock/;
use dag::grammar;

require Exporter;

our @ISA = qw(Exporter);
our $typeof;
our %EXPORT_TAGS = ( 'all' => [ qw() ] );
our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );
our @EXPORT = qw($typeof %ST);

our %ST;
our $VERSION = '0.02';

# this variable is used for the tokenizer
our $rt = undef;

### Tokenize (yylex)
#
sub tokenize {
  my $text = shift;

  $text = $rt if defined $rt;
  $text =~ s!^\s*!!;

  my $word = '(\!?(\.\w|\w\d+|\w|<\w+>)+)';
  for ($text) {
    if (/^$word/)                 { $rt = $'; return ('AGOD_ID', $&); }
    if (/^--->/)                  { $rt = $'; return ('AGOD_ARROW', $&); }
    if (/^[*+?]/)                 { $rt = $'; return ('AGOD_QUANTIFIER', $&); }
    if (/^::/)                    { $rt = $'; return ('AGOD_FOUR_DOTS', $&); }
    if (/^(\[|\])/)               { $rt = $'; return ($&, $&); }
    if (/^do\b/)                  { $rt = $'; return ('&','&'); }
    if (/^[,;&=]/)                { $rt = $'; return ($&, $&); }
    if (/^%%/)                    { $rt = ""; return ('AGOD_FINAL', $'); }
    if (/^(\{(\n|.)*?\})(\s|\n)*\n(%%|::|\s*($word\s*(--->|;(\s|\n)*($word\s*;(\s|\n)*)*(::|%%))))/) {
      ### WEIRD, but works
      $rt = $4.$';
      return ('AGOD_PERL', $1);
    }
    $rt = $';
    return undef
  }
}

### yyerror for parser
#
sub yyerror {
  my $msg =shift;
  die ("error: $msg ($rt)");
}

sub complete {
  my $grammar = shift;
  for my $simb (keys %{$grammar}) {
    my $string = "\$_ = shift;\nmy \%ME = \%{\$_}; \n";
    my %rhs = %{$grammar->{$simb}{'%RHS'}};
    for my $s (sort keys %rhs) {
      if ($rhs{$s} eq "*" or $rhs{$s} eq "+") {
	$string .= 'my $'.$s.' = shift;'."\n";
	$string .= 'my @'.$s.' = defined($'.$s.')?@{$'.$s.'}:();'."\n";
      } elsif ($rhs{$s} eq "1" or $rhs{$s} eq "?") {
	$string .= 'my $'.$s.' = shift;'."\n";
      } else {
	die("Grammar corrupted");
      }
    }
    if (defined($grammar->{$simb}->{'%BEGIN'})) {
      $grammar->{$simb}{'%BEGIN'} = 'XXXDAGXXX sub { '.
	$string.$grammar->{$simb}{'%BEGIN'}."}XXXDAGXXX";
    }
    for my $att (grep {/^[^%]/} keys %{$grammar->{$simb}}) {
      my $tmp;
      if (defined($tmp = $grammar->{$simb}{$att}{RETURNS})) {
	$grammar->{$simb}{$att}{RETURNS} = 'XXXDAGXXX sub { '.$string.$tmp."}XXXDAGXXX";
      } elsif (defined($tmp = $grammar->{$simb}{$att}{EXECUTE})) {
	$grammar->{$simb}{$att}{EXECUTE} = 'XXXDAGXXX sub { '.$string.$tmp."}XXXDAGXXX";
      }
    }
  }
}

sub parse {
  # Receives all the grammar file by first argument
  # So far, ignores all code following the last %%.
  my $grammar = shift;
  $grammar =~ s/\s*%%//;

  $rt = ($grammar =~ /%%\s*$/)?$grammar:$grammar."\n%%";

  my $parser = new dag::grammar;

  return bless($parser->YYParse(yylex => \&tokenize,
				yyerror => \&yyerror)),
}


### uptodate(file, f1, f2, ... fn) ####
#                                     #
# Returns 0 is file is newer than f_i #
#                                     #
#######################################
sub uptodate {
  my $file = shift;
  return 0 unless (-e $file);
  for (@_) {
    return 1 if (-e $_ && -M $file > -M $_)
  }
  return 0;
}

sub non_uptodate {
  my $file = shift;
  return 1 unless (-e $file);
  for (@_) {
    return 1 if (-e $_ && -M $file > -M $_)
  }
  return 0;
}


sub notuptodate {
  my $file = shift;
  return 0 unless (-e $file);
  for (@_) {
    return 1 if (-e $_ && -M $file < -M $_);
    return 1 unless (-e $_);
  }
  return 0;
}


### This function should be defined by the user, but for testing
#   purpose, I'll define it here.
#
#   Receives the filename and the path to the file.
#
#   Should return a string like the one used in the grammar
#
#   we should ensure that the function is called at the correct path
#
##################################################################
sub typeof {
  return &{$typeof}(@_);
}

sub getfile {
  my $fullname = shift;
  $fullname =~ m!/([^/]+)/?$!;
  return $1;
}

sub getdirectory {
  my $fullname = shift;
  $fullname =~ m!/([^/]+)/?$!;
  return $`;
}

sub error_die {
  error(@_);
  die;
}

# Prints error messages
sub error {
  my ($dir,$message) = @_;
  $dir = "...".substr($dir,-65);
  print " ** ERROR at $dir **\n";
  print " ** $message\n";
}

# Given a grammar and a type, returns the right side hash table
sub RHS {
  my ($grammar,$type) = @_;
  return undef unless defined($grammar->{$type});
  return %{$grammar->{$type}->{'%RHS'}};
}

# returns the list of _FILES_ of the current directory
sub files {
  return grep {/^[^_]/} map {chomp; $_} `ls -1`
}

sub analyse {
  my ($grammar,$dir) = @_;
  my $type = typeof(getfile($dir),getdirectory($dir));
  my $itsadir = 0;
  if (defined($grammar->{$type})) {
    # Save old directory...
    chomp(my $oldir = `pwd`);

    # Initialize symbol table;
    %ST = ();

    my @attrs = ();

    if (-d $dir) {
      $itsadir = 1;
      chdir($dir);

      # get directory files...
      my @files = files();
      for (@files) {
	  my $r = $grammar -> analyse("$dir/$_");
	  push @attrs, $r if defined $r->[1];
      }
    }

    my $underscore = { _ => getfile($dir) };

    # Process the BEGIN directive
    if (defined($grammar->{$type}->{'%BEGIN'})) {
      $grammar->{$type}{'%BEGIN'}($underscore);
    }

    # Creates variables... but doesn't check grammar right hand side
    my @vars = ();
    if (my %rhs = RHS($grammar,$type)) {
      for my $v (sort keys %rhs) {
	if    ($rhs{$v} eq '?' || $rhs{$v} eq '1') {
	  my @rs = map {$_->[1]} grep {$_->[0] eq $v} @attrs;
	  if (@rs) {
	    push @vars, $rs[0];
	  } else {
	    push @vars, "";
	  }
	}
	elsif ($rhs{$v} eq '*' || $rhs{$v} eq '+') {
	  my @x =  map {$_->[1]} grep {$_->[0] eq $v}@attrs;
	  push @vars, [@x];
	}
	else  { error($dir,"Unkown quantifier for $rhs{$_}") }
      }
    }

    if (-d $dir) {
      chdir ($oldir);
    }

    # Process all attributes
    my @attodo = $grammar->get_atts($type);
    for my $attribute (@attodo) {
      # Prepare variables
	$underscore->{$attribute} = attr($attribute, $underscore->{_},%ST);
	$_ = $underscore;

	my $str = "Calculating $_->{_} to $_->{$attribute} ";
	if (length($str)>70) {
	  print substr(0,70);
	} else {
	  print $str;
	  print "." x (70-length($str));
	}

	# my @fs = map {my %h = %{$_->[1]}; values %h} @attrs;
	my @fs = map {values %{$_->[1]}} @attrs;

      #Check reccentness
      if ((-d $dir && non_uptodate($_->{$attribute},@fs)) ||
	  notuptodate($_->{_},$_->{$attribute})) {
	  if (defined($grammar->{$type}->{$attribute}->{RETURNS})) {
	      chdir($dir) if -d $dir;

	      my $retval = $grammar->{$type}{$attribute}{RETURNS}($underscore,@vars);
	      chdir($oldir) if -d $dir;
	      error_die($dir,$@) if $@;

	      open OUTPUT, ">$underscore->{$attribute}" or error_die($dir, $!);
	      print OUTPUT $retval;
	      close OUTPUT;

	  } elsif (defined($grammar->{$type}->{$attribute}->{EXECUTE})) {
	      chdir($dir) if -d $dir;
	      $grammar->{$type}{$attribute}{EXECUTE}($underscore,@vars);
	      chdir($oldir) if -d $dir;
	      error_die($dir,$@) if $@;
	  }
	  print "DONE\n";
      } else {
	  print "UPTODATE\n";
      }

    #my $x = getfile(getdirectory($dir));
    #for (keys %{$underscore}) {
    #$underscore->{$_} = "$x/$underscore->{$_}";
    #}
    }
    return [$type,$underscore];
  } else {
    error($dir,"Type unkown: '$type'. Bailing out!");
    return [$type,undef];
  }
}

### First argument is the grammar object followed by the directory
#   grammar root (a simple directory)
#
########################################################################
sub run {
  my $self = shift;
  my $grammar = $self;
  my $directory = shift;
  $| = 1;

  unless ($directory =~ m!^/!) {
    chomp(my $d = `pwd`);
    $directory = "$d/$directory";
  }

  $grammar -> analyse($directory);
  return;

  chomp(my $oldir = `pwd`);
  chdir($directory) or die("Cannot chdir to $directory ($!)\n");

  checkfiles($grammar,$directory);

  my @files = grep {/^[^_]/} map {chomp; $_} `ls -1`;

  my @dirs  = grep {-d "$directory/$_" } @files;
  @files = grep { -f "$directory/$_" } @files;

  for (@dirs) {
    run($self, "$directory/$_");
  }
  for (@files) {
    process($grammar,$directory,$_);
  }
  process($grammar,$directory);

  chdir($oldir) or die("Cannot chdir to $oldir ($!)\n");
}

### Does the job? :)
sub process {
  my ($grammar, $dir, $file) = @_;
  my $type;

  # This is a file, process it...
  if ($file) {
    $type = typeof($file,$dir);

    if (defined($grammar->{$type})) {
      my $tmp = "$dir/$file";
      $tmp = "...".substr($tmp,-50) if length($tmp)>50;
      print "Processing $tmp ($type)\n";
      my @todo = get_atts($grammar,$type);

      my $att;
      my $expression = '$_ = {_=>"'.$file.'"};';

      # Prepare symbol table
      %ST = ();

      # Evaluate BEGIN block
      eval("$expression ".$grammar->{$type}->{'%BEGIN'}) if defined ($grammar->{$type}->{'%BEGIN'});
      warn("Error\@$type begin sub: $@") if $@;

      # While there are attributes to be calculated
      while($att = shift @todo) {
	print "\tcalculating attribute ::$att...";

	### [TODO] Check uptodate files
	if (uptodate($file,attr($att,$file,%ST))) {
	  print " uptodate\n";
	} else {
	  ### [TODO] Check dependencies

	  # Set variable file name
	  $expression.= '$_->{"'.$att.'"}="'.attr($att,$file,%ST).'";';
	  calculate_attr($grammar,$type,$att,$expression,$dir,%ST);
	  print " done\n";
	}
      }
    } else {
      # print STDERR "Grammar does not says anything about type $type\n";
    }
  } else {
    # Get directory type
    chdir(getdirectory($dir));
    $type = typeof(getfile($dir),getdirectory($dir));
    chdir($dir);

    # Directory

    if (defined($grammar->{$type})) {
      my $tmp = "$dir/";
      $tmp = "...".substr($tmp,-50) if length($tmp)>50;
      print "Processing $tmp ($type)\n";

      # Check file types
      my @files = grep {/^[^_]/} map {chomp; $_} `ls -1`;
      my %rhs;
      for (map {[$_,typeof($_,$dir)]} @files) {
	push @{$rhs{$_->[1]}}, $_->[0];
      }
      # Calculate dependent attributes
      my %RHS;
      for my $t(keys %rhs) {
	if (defined($grammar->{$t})) {
	  for my $f(@{$rhs{$t}}) {
	    my @atts = get_atts($grammar,$t);
	    my $x = {_=>$f};
	    for (@atts) {
	      $x->{$_} = attr($_,$f);
	    }
	    push @{$RHS{$t}}, $x;
	  }
	}
      }

      # Prepare variables
      my $expression = '$_={_=>"'.getfile($dir).'"};';

      # Prepare symbol table
      %ST = ();
      eval($expression . $grammar->{$type}->{'%BEGIN'}) if defined ($grammar->{$type}->{'%BEGIN'});
      warn("Error\@$type begin sub: $@") if $@;

      for my $t (keys %RHS) {
	if (defined($grammar->{$type}->{'%RHS'}->{$t})) {
	  my $quantifier = $grammar->{$type}->{'%RHS'}->{$t};
	  if ($quantifier eq "*" || $quantifier eq "+") {
	    $expression.='my @'.$t.'=('.
	      join(",",map { my_dump($_) } @{$RHS{$t}}).
		');';
	  } elsif ($quantifier eq '1' || $quantifier eq '?') {
	    $expression.='my $'.$t.'='.my_dump($RHS{$t}[0]).';';
	  } else {
	    warn("Error\@$dir: strange quantifier '$quantifier' for ($type,$t)");
	  }
	} else {
	  # warn("Error\@$dir: $t should not be under $type");
	}
      }
      # Check for '?' and '*' quantifiers missing from current dir
      my @t = keys %{$grammar->{$type}->{'%RHS'}};
      for (@t) {
	next if defined($RHS{$_});
	if ($grammar->{$type}->{'%RHS'}->{$_} eq '?') {
	  $expression.='my $'.$_.'=undef;';
	} elsif ($grammar->{$type}->{'%RHS'}->{$_} eq '*') {
	  $expression.='my @'.$_.'=();';
	}
      }

      # Calculate each attribute
      my @todo = get_atts($grammar,$type);
      my $att;
      while($att = shift @todo) {
	print "\tcalculating attribute ::$att...";

	if (0) { #!uptodate("../".attr($att,getfile($dir),%ST), @files)) {
	  print " uptodate\n";
	} else {
	  ### [TODO] Check dependencies

	  # Set variable file name
	  $expression.= '$_->{"'.$att.'"}="'.attr($att,"../".getfile($dir),%ST).'";';
	  calculate_attr($grammar,$type,$att,$expression,$dir,%ST);
	  print " done\n";
	}
      }
    }
    else {
      # print STDERR "Type $type does not exists in the grammar\n";
    }
  }
}

sub checkfiles {
  my ($grammar,$cdir) = @_;

  my $dpath = getdirectory($cdir);
  chdir($dpath);
  my $type = typeof(getfile($cdir),$dpath);
  chdir($cdir);

  unless (defined ($grammar->{$type})) {
    warn ("ERROR\@$cdir: unknown type $type");
    return;
  }

  my @types = map {chomp; typeof($_,$cdir)} `ls -1`;
  my %types;
  for (@types) {
    $types{$_}++;
  }

  my %real_types = %{$grammar->{$type}->{'%RHS'}};

  # Check for types that shouldn't be here...
  for (keys %types) {
    unless (defined($real_types{$_})) {
      # warn ("ERROR\@$cdir: there is a type $_ under a type $type");
    }
  }

  # Check for number of types
  for (keys %real_types) {
    if ($real_types{$_} eq "1") {
      warn ("ERROR\@$cdir: there are too many types $_") unless (defined($types{$_}) and $types{$_} == 1);
    } elsif ($real_types{$_} eq "*") {
      # Do nothing, so far
    } elsif ($real_types{$_} eq "+") {
      warn ("ERROR\@$cdir: there should be at least one type $_") unless defined($types{$_});
    } elsif ($real_types{$_} eq "?") {
      warn ("ERROR\@$cdir: there are too many types $_") unless (!defined($types{$_}) || (defined($types{$_}) && $types{$_} == 1));
    }
  }
}

sub attr {
  my $attribute = shift;
  my $file = shift;

  %ST = @_;

  $attribute =~ s!<([^.]+)>!defined($ST{$1})?$ST{$1}:""!ge;

  my $prefix = "";
  if ($attribute =~ /^!/) {
    my $f = $';
    if (-d $file) {
      mkdir("$file/_") unless -d "$file/_";
      "$file/_/$f"
    } else {
      mkdir("_$file") unless -d "_$file";
      "_$file/$f"
    }
  } elsif (-d $file) {
    if ($attribute =~ /(.*)\.(.*)/) { "$file/_$1_.$2" }
    else                            { "$file/_${attribute}_" }
  } else {
    if($attribute =~/(.*)\.(.*)/)  { "_$1_$file.$2"  }
    else                           { "_${attribute}_${file}"  }
  }

}


sub RETURNS {
  my ($grammar,$type,$att) = @_;
  my $expression = 'open F, ">".$_->{"'.$att.'"};';
  $expression .= 'my $t = sub'.$grammar->{$type}->{$att}->{RETURNS}.';';
  $expression .= 'print F &{$t};';
  $expression .= 'close F;';
}

sub EXECUTE {
  my ($grammar,$type,$att) = @_;
  my $expression = 'my $t = sub'.$grammar->{$type}->{$att}->{EXECUTE}.';';
  $expression .= '&{$t};';
}

sub get_atts {
  my ($grammar,$type) = @_;
  return grep {/^[^%]/} keys %{$grammar->{$type}};
}

sub my_dump {
  my $o = Dumper(shift);
  $o =~ s/^.*?\{/\{/;
  $o =~ s/\};\n/\}/;
  $o
}


sub calculate_attr {
  my ($grammar,$type,$att,$expression,$dir,%ST) = @_;
  if (defined($grammar->{$type}->{$att}->{RETURNS})) {
    my $exp = $expression . RETURNS($grammar,$type,$att);
    eval $exp;
    die "ERROR\@$dir/ $@\n" if ($@);
  } elsif (defined($grammar->{$type}->{$att}->{EXECUTE})) {
    my $exp = $expression . EXECUTE($grammar,$type,$att);
    eval $exp;
    die "ERROR\@$dir/ $@\n" if ($@);
  } else {
    warn "Error calculating attribute $att\@$dir"
  }
}


sub contents {
  my $par = shift;
  return "" unless defined $par;
  my $arg;
  if ($arg = shift) {
    return `cat $par->{$arg}`;
  } else {
    return `cat $par->{_}`;
  }
}


1;
__END__

=head1 NAME

dag - Perl extension for atribute grammars over directory

=head1 SYNOPSIS

  use dag;


While we can look to a directory as a trash box, it can be defined a
formal grammar defining the way it would be organized.

=head1 AUTHOR

Jos Joo Almeida <jj@di.uminho.pt>

Alberto M. Simes <albie@alfarrabio.di.uminho.pt>

=head1 SEE ALSO

perl(1).

=cut
